"""A set of special exceptions that can be thrown by the jplephem library"""

class OutOfRangeError(ValueError):
    """One or more time values given were out of range for the ephemeris.

    This exception is thrown if any input times are out of the range of
    times supported by an ephemeris.  It has an extra attribute:

    - `out_of_range_times`: if the input `tdb` of times is an array,
      this provides an array of booleans of the same length where `True`
      means the corresponding date is out of range.

    """
    def __init__(self, message, out_of_range_times):
        self.args = message,
        self.out_of_range_times = out_of_range_times



"""Routines for dealing with Julian dates."""

def compute_calendar_date(jd_integer, julian_before=None):
    """Convert Julian day ``jd_integer`` to ``(year, month, day)``.

    Uses the proleptic Gregorian calendar unless ``julian_before`` is
    set to a specific Julian day, in which case the Julian calendar is
    used for dates older than that.

    """
    use_gregorian = (julian_before is None) or (jd_integer >= julian_before)

    # See the Explanatory Supplement to the Astronomical Almanac 15.11.
    f = jd_integer + 1401
    f += use_gregorian * ((4 * jd_integer + 274277) // 146097 * 3 // 4 - 38)
    e = 4 * f + 3
    g = e % 1461 // 4
    h = 5 * g + 2
    day = h % 153 // 5 + 1
    month = (h // 153 + 2) % 12 + 1
    year = e // 1461 - 4716 + (12 + 2 - month) // 12
    return year, month, day

def compute_julian_date(year, month=1, day=1.0):
    """Given a proleptic Gregorian date, return a Julian date float."""
    return compute_julian_day(year, month, day) - 0.5

def compute_julian_day(year, month=1, day=1):
    """Given a proleptic Gregorian date, return a Julian day int."""
    janfeb = month < 3
    return (+ 1461 * (year + 4800 - janfeb) // 4
            + 367 * (month - 2 + janfeb * 12) // 12
            - 3 * ((year + 4900 - janfeb) // 100) // 4
            - 32075
            + day)



"""Compute positions from an ephemeris installed as a Python package.

Note: This entire module is DEPRECATED.  The idea of distributing JPL
ephemerides as Python packages proved to be impractical (they were much
too large for the Python Package Index to easily store and distribute),
and it forced `jplephem` users get their ephemerides from a different
source than mainline astronomers, who use SPICE files.

"""
import os
import numpy as np

class DateError(ValueError):
    pass

class Ephemeris(object):
    """[DEPRECATED] JPL planetary ephemeris for computing positions on dates."""

    def __init__(self, module):
        self.name = module.__name__.upper()
        self.dirpath = os.path.dirname(module.__file__)
        self.names = tuple(sorted(
            name.split('-')[-1].split('.')[0]
            for name in os.listdir(self.dirpath)
            if not name.startswith('constants') and name.endswith('.npy')
            ))
        path = self.path('constants.npy')
        self.__dict__.update((k.decode('ascii'), v) for k, v in np.load(path))
        self.earth_share = 1.0 / (1.0 + self.EMRAT)
        self.moon_share = self.EMRAT / (1.0 + self.EMRAT)
        self.sets = {}

    def path(self, filename):
        """[DEPRECATED] Compute the path to a particular file in the ephemeris."""
        return os.path.join(self.dirpath, filename)

    def load(self, name):
        """[DEPRECATED] Load the polynomial series for `name` and return it."""
        s = self.sets.get(name)
        if s is None:
            self.sets[name] = s = np.load(self.path('jpl-%s.npy' % name))
        return s

    def position(self, name, tdb, tdb2=0.0):
        """[DEPRECATED] Compute the position of `name` at ``tdb [+ tdb2]``."""
        bundle = self.compute_bundle(name, tdb, tdb2)
        return self.position_from_bundle(bundle)

    def position_and_velocity(self, name, tdb, tdb2=0.0):
        """[DEPRECATED] Compute the position and velocity of `name` at ``tdb [+ tdb2]``."""
        bundle = self.compute_bundle(name, tdb, tdb2)
        position = self.position_from_bundle(bundle)
        velocity = self.velocity_from_bundle(bundle)
        return position, velocity

    def compute(self, name, tdb):
        """[DEPRECATED] Legacy routine that concatenates position and velocity vectors."""
        bundle = self.compute_bundle(name, tdb, 0.0)
        position = self.position_from_bundle(bundle)
        velocity = self.velocity_from_bundle(bundle)
        return np.concatenate((position, velocity))

    def compute_bundle(self, name, tdb, tdb2=0.0):
        """[DEPRECATED] Return a tuple of coefficients and parameters for `tdb`."""
        input_was_scalar = getattr(tdb, 'shape', ()) == ()
        if input_was_scalar:
            tdb = np.array((tdb,))
        # no need to deal with tdb2; numpy broadcast will add fine below.

        coefficient_sets = self.load(name)
        number_of_sets, axis_count, coefficient_count = coefficient_sets.shape

        jalpha, jomega = self.jalpha, self.jomega
        days_per_set = (jomega - jalpha) / number_of_sets
        # to keep precision, first subtract, then add
        index, offset = divmod((tdb - jalpha) + tdb2, days_per_set)
        index = index.astype(int)

        if (index < 0).any() or (number_of_sets < index).any():
            raise DateError('ephemeris %s only covers dates %.1f through %.1f'
                            % (self.name, jalpha, jomega))

        omegas = (index == number_of_sets)
        index[omegas] -= 1
        offset[omegas] += days_per_set

        coefficients = np.rollaxis(coefficient_sets[index], 1)

        # Chebyshev recurrence:

        T = np.empty((coefficient_count, len(index)))
        T[0] = 1.0
        T[1] = t1 = 2.0 * offset / days_per_set - 1.0
        twot1 = t1 + t1
        for i in range(2, coefficient_count):
            T[i] = twot1 * T[i-1] - T[i-2]

        bundle = coefficients, days_per_set, T, twot1
        return bundle

    def position_from_bundle(self, bundle):
        """[DEPRECATED] Return position, given the `coefficient_bundle()` return value."""

        coefficients, days_per_set, T, twot1 = bundle
        return (T.T * coefficients).sum(axis=2)

    def velocity_from_bundle(self, bundle):
        """[DEPRECATED] Return velocity, given the `coefficient_bundle()` return value."""

        coefficients, days_per_set, T, twot1 = bundle
        coefficient_count = coefficients.shape[2]

        # Chebyshev derivative:

        dT = np.empty_like(T)
        dT[0] = 0.0
        dT[1] = 1.0
        dT[2] = twot1 + twot1
        for i in range(3, coefficient_count):
            dT[i] = twot1 * dT[i-1] - dT[i-2] + T[i-1] + T[i-1]
        dT *= 2.0
        dT /= days_per_set

        return (dT.T * coefficients).sum(axis=2)



"""Tests for ``jplephem``.

See the accompanying ``jpltest`` module for a more intense numerical
test suite that can verify that ``jplephem`` delivers, over hundreds of
examples, the same results as when the ephemerides are run at JPL.  This
smaller and more feature-oriented suite can be run with::

    python -m unittest discover jplephem

"""
import gc
import mmap
import numpy as np
import sys
import tempfile
import warnings
from doctest import DocTestSuite, ELLIPSIS
from functools import partial
from io import BytesIO
from jplephem import Ephemeris, commandline
from jplephem.exceptions import OutOfRangeError
from jplephem.daf import DAF, FTPSTR, NAIF_DAF
from jplephem.pck import PCK
from jplephem.spk import SPK
from struct import Struct
try:
    from unittest import SkipTest, TestCase
except ImportError:
    from unittest2 import SkipTest, TestCase

epsilon_m = 0.01
target_names = {
    'mercury barycenter': 1, # BARYCENTER w.r.t. 0 SOLAR SYSTEM BARYCENTER
    'venus barycenter': 2, # BARYCENTER w.r.t. 0 SOLAR SYSTEM BARYCENTER
    'earthmoon': 3,        # BARYCENTER w.r.t. 0 SOLAR SYSTEM BARYCENTER
    'mars barycenter': 4,  # BARYCENTER w.r.t. 0 SOLAR SYSTEM BARYCENTER
    'jupiter': 5,          # BARYCENTER w.r.t. 0 SOLAR SYSTEM BARYCENTER
    'saturn': 6,           # BARYCENTER w.r.t. 0 SOLAR SYSTEM BARYCENTER
    'uranus': 7,           # BARYCENTER w.r.t. 0 SOLAR SYSTEM BARYCENTER
    'neptune': 8,          # BARYCENTER w.r.t. 0 SOLAR SYSTEM BARYCENTER
    'pluto': 9,            # BARYCENTER w.r.t. 0 SOLAR SYSTEM BARYCENTER
    'sun': 10,             # w.r.t. 0 SOLAR SYSTEM BARYCENTER
    'mercury': 199,        # w.r.t. 1 MERCURY BARYCENTER
    'venus': 299,          # w.r.t. 2 VENUS BARYCENTER
    'moon': 301,           # w.r.t. 3 EARTH BARYCENTER
    'earth': 399,          # w.r.t. 3 EARTH BARYCENTER
    'mars': 499,           # w.r.t. 4 MARS BARYCENTER
    }

class TestDAFBytesIO(TestCase):
    def sample_daf(self):
        word = Struct('<d').pack
        integer = Struct('<i').pack
        return BytesIO(b''.join([
            # Record 1 - File Record
            b'DAF/SPK ',
            b'\x02\x00\x00\x00', # ND
            b'\x03\x00\x00\x00', # NI
            b'Internal Name'.ljust(60, b' '), # LOCIFN
            b'\x03\x00\x00\x00', # FWARD
            b'\x07\x00\x00\x00', # BWARD
            b'\x01\x04\x00\x00', # FREE
            b'LTL-IEEE', # LOCFMT
            b'\0' * 603, # PRENUL
            FTPSTR,
            b'\0' * 297, # PSTNUL

            # Record 2
            b'Comment Record'.ljust(1024, b'\0'),

            # Record 3 - first Summary Record
            b''.join([
                word(7), # next summary record
                word(0), # previous summary record
                word(1), # number of summaries
                word(101),
                word(202),
                integer(303),
                integer(1024 * 4 // 8 + 1), # Record 5 start
                integer(1024 * 5 // 8),     # Record 5 end
                integer(0),
            ]).ljust(1024, b'\0'),

            # Record 4 - first Name Record
            b'Summary Name 1'.ljust(1024, b' '),

            # Record 5
            word(1001) * 128,

            # Record 6
            word(2002) * 128,

            # Record 7 - second Summary Record
            b''.join([
                word(0), # next summary record
                word(3), # previous summary record
                word(1), # number of summaries
                word(111),
                word(222),
                integer(333),
                integer(1024 * 5 // 8 + 1), # Record 6 start
                integer(1024 * 6 // 8),     # Record 6 end
                integer(0),
            ]).ljust(1024, b'\0'),

            # Record 8 - second Name Record
            b'Summary Name 2'.ljust(1024, b' '),
        ]))

    def test_header(self):
        f = self.sample_daf()
        d = DAF(f)
        eq = self.assertEqual
        eq(d.locidw, b'DAF/SPK')
        eq(d.nd, 2)
        eq(d.ni, 3)
        eq(d.locifn_text, b'Internal Name')
        eq(d.fward, 3)
        eq(d.bward, 7)
        eq(d.free, 0x401)
        eq(d.locfmt, b'LTL-IEEE')

    def test_segments(self):
        f = self.sample_daf()
        d = DAF(f)

        summaries = list(d.summaries())
        eq = self.assertEqual
        eq(len(summaries), 2)
        eq(summaries[0], (b'Summary Name 1', (101.0, 202.0, 303, 513, 640)))
        eq(summaries[1], (b'Summary Name 2', (111.0, 222.0, 333, 641, 768)))

        eq = self.assertSequenceEqual
        eq(list(d.map(summaries[0][1])), [1001.0] * 128)
        eq(list(d.map(summaries[1][1])), [2002.0] * 128)

    def test_add_segment(self):
        f = self.sample_daf()
        d = DAF(f)

        d.add_array(b'Summary Name 3', (121.0, 232.0, 343), [3003.0] * 128)

        summaries = list(d.summaries())
        eq = self.assertEqual
        eq(len(summaries), 3)
        eq(summaries[0], (b'Summary Name 1', (101.0, 202.0, 303, 513, 640)))
        eq(summaries[1], (b'Summary Name 2', (111.0, 222.0, 333, 641, 768)))
        eq(summaries[2], (b'Summary Name 3', (121.0, 232.0, 343, 1025, 1152)))

        eq = self.assertSequenceEqual
        eq(list(d.map(summaries[0][1])), [1001.0] * 128)
        eq(list(d.map(summaries[1][1])), [2002.0] * 128)
        eq(list(d.map(summaries[2][1])), [3003.0] * 128)

    def test_add_segment_when_summary_block_is_full(self):
        f = self.sample_daf()
        d = DAF(f)

        # Update n_summaries of final summary block to full.
        d.file.seek(6 * 1024 + 16)
        d.file.write(Struct('<d').pack(d.summaries_per_record))

        d.add_array(b'Summary Name 3', (121.0, 232.0, 343), [3003.0] * 200)

        # Reset n_summaries of that block back to its real value.
        d.file.seek(6 * 1024 + 16)
        d.file.write(Struct('<d').pack(1))

        summaries = list(d.summaries())
        eq = self.assertEqual
        eq(len(summaries), 3)
        eq(summaries[0], (b'Summary Name 1', (101.0, 202.0, 303, 513, 640)))
        eq(summaries[1], (b'Summary Name 2', (111.0, 222.0, 333, 641, 768)))
        eq(summaries[2], (b'Summary Name 3', (121.0, 232.0, 343, 1281, 1480)))

        eq = self.assertSequenceEqual
        eq(list(d.map(summaries[0][1])), [1001.0] * 128)
        eq(list(d.map(summaries[1][1])), [2002.0] * 128)
        eq(list(d.map(summaries[2][1])), [3003.0] * 200)

class TestDAFRealFile(TestDAFBytesIO):
    # Where "Real" = "written to disk with a real file descriptor
    # instead of an in-memory BytesIO".

    def sample_daf(self):
        bytes_io = super(TestDAFRealFile, self).sample_daf()
        f = tempfile.NamedTemporaryFile(mode='w+b', prefix='jplephem_test')
        f.write(bytes_io.getvalue())
        f.seek(0)
        return f

def fake_mmap_that_raises_OSError(*args, **kw):
    raise OSError('mmap() not supported on this platform')

class TestDAFRealFileWithoutMMap(TestDAFRealFile):
    # Where "Real" = "written to disk with a real file descriptor
    # instead of an in-memory BytesIO".  And we turn off mmap() to
    # simulate platforms like pyodide.

    def setUp(self):
        self.mmap = mmap.mmap
        mmap.mmap = fake_mmap_that_raises_OSError

    def tearDown(self):
        mmap.mmap = self.mmap

class _CommonTests(object):

    def check0(self, xyz, xyzdot=None):
        eq = partial(self.assertAlmostEqual, delta=epsilon_m)
        x, y, z = xyz
        eq(x, 39705023.28)
        eq(y, 131195345.65)
        eq(z, 56898495.41)
        if xyzdot is None:
            return
        dx, dy, dz = xyzdot
        eq(dx, -2524248.19)
        eq(dy, 619970.11)
        eq(dz, 268928.26)

    def check1(self, xyz, xyzdot=None):
        eq = partial(self.assertAlmostEqual, delta=epsilon_m)
        x, y, z = xyz
        eq(x, -144692624.00)
        eq(y, -32707965.14)
        eq(z, -14207167.26)
        if xyzdot is None:
            return
        dx, dy, dz = xyzdot
        eq(dx, 587334.38)
        eq(dy, -2297419.36)
        eq(dz, -996628.74)

    def test_scalar_tdb(self):
        self.check0(self.position('earthmoon', 2414994.0))
        self.check1(self.position('earthmoon', 2415112.5))

    def test_scalar_tdb2(self):
        self.check0(self.position('earthmoon', 2414990.0, 4.0))
        self.check1(self.position('earthmoon', 2415110.0, 2.5))

    def test_scalar_tdb_keyword(self):
        self.check0(self.position('earthmoon', tdb=2414994.0))
        self.check1(self.position('earthmoon', tdb=2415112.5))

    def test_scalar_tdb2_keyword(self):
        self.check0(self.position('earthmoon', tdb=2414990.0, tdb2=4.0))
        self.check1(self.position('earthmoon', tdb=2415110.0, tdb2=2.5))

    def check_2d_result(self, name, tdb, tdb2):
        p = self.position(name, tdb + tdb2)
        self.check0(p[:,0])
        self.check1(p[:,1])

        p = self.position(name, tdb, tdb2)
        self.check0(p[:,0])
        self.check1(p[:,1])

        p, v = self.position_and_velocity(name, tdb + tdb2)
        self.check0(p[:,0], v[:,0])
        self.check1(p[:,1], v[:,1])

        p, v = self.position_and_velocity(name, tdb, tdb2)
        self.check0(p[:,0], v[:,0])
        self.check1(p[:,1], v[:,1])

    def test_array_tdb(self):
        tdb = np.array([2414994.0, 2415112.5])
        tdb2 = 0.0
        self.check_2d_result('earthmoon', tdb, tdb2)

    def test_array_tdb_scalar_tdb2(self):
        tdb = np.array([2414991.5, 2415110.0])
        tdb2 = 2.5
        self.check_2d_result('earthmoon', tdb, tdb2)

    def test_scalar_tdb_array_tdb2(self):
        tdb = 2414990.0
        d = 2415112.5 - tdb
        tdb2 = np.array([4.0, d])
        self.check_2d_result('earthmoon', tdb, tdb2)

    def test_array_tdb_array_tdb2(self):
        tdb = np.array([2414990.0, 2415110.0])
        tdb2 = np.array([4.0, 2.5])
        self.check_2d_result('earthmoon', tdb, tdb2)

    def test_jitter(self):
        usecond = 1e-6 / 24.0 / 3600.0
        tdb = np.ones(12) * 2414998.0
        tdb2 = np.linspace(1.0 * usecond, 2.0 * usecond, 12)
        x, y, z = self.position('earthmoon', tdb, tdb2)
        for component in x, y, z:
            size = component[0]
            relative_jitter = np.diff(np.diff(x)) / size
            self.assertLess(max(abs(relative_jitter)), 3e-16)

    def test_ephemeris_end_date(self):
        x, y, z = self.position('earthmoon', self.jomega)
        # These positions are actually from HORIZONS and thus DE431,
        # hence the low precision match:
        self.assertAlmostEqual(x, 1.442502234663646E+08, delta=1.0)
        self.assertAlmostEqual(y, 3.690043031712407E+07, delta=1.0)
        self.assertAlmostEqual(z, 1.599543968176661E+07, delta=1.0)

    def test_too_early_date(self):
        tdb = self.jalpha - 0.01
        self.assertRaises(ValueError, self.position, 'earthmoon', tdb)

    def test_too_late_date(self):
        tdb = self.jomega + 16.01
        self.assertRaises(ValueError, self.position, 'earthmoon', tdb)

class SPKTests(_CommonTests, TestCase):

    def setUp(self):
        try:
            self.spk = SPK.open('de421.bsp')
        except IOError:
            raise SkipTest('the "de421.bsp" SPK file is not available')
        segment = self.spk[0,1]
        self.jalpha = segment.start_jd
        self.jomega = segment.end_jd

    def tearDown(self):
        self.spk.close()

        if sys.version_info < (3,):
            return

        # With thanks for https://stackoverflow.com/questions/24717027/
        with warnings.catch_warnings(record=True) as w:
            warnings.resetwarnings()
            warnings.simplefilter('always', ResourceWarning)
            del self.spk
            gc.collect()
            self.assertFalse(w and str(w[-1]))

    def position(self, name, tdb, tdb2=0.0):
        segment = self.spk[0, target_names[name]]
        return segment.compute(tdb, tdb2)

    def position_and_velocity(self, name, tdb, tdb2=0.0):
        segment = self.spk[0, target_names[name]]
        return segment.compute_and_differentiate(tdb, tdb2)

    def test_segment_with_only_two_coefficients(self):
        tdb = 2414990.0
        tup = target_names['mercury barycenter'], target_names['mercury']
        segment = self.spk[tup]
        segment.compute_and_differentiate(tdb)

    def test_str(self):
        str(self.spk)  # just to confirm it does not raise an exception
        segment = self.spk[0,4]
        self.assertEqual(str(segment), segment.describe(verbose=False))
        self.assertEqual(segment.describe(verbose=False),
  '1899-07-29..2053-10-09  Type 2  Solar System Barycenter (0) -> Mars Barycenter (4)')
        self.assertEqual(segment.describe(verbose=True),
  '1899-07-29..2053-10-09  Type 2  Solar System Barycenter (0) -> Mars Barycenter (4)'
  '\n  frame=1 source=DE-0421LE-0421')

    def test_loading_array(self):
        segment = self.spk[0,4]
        initial_epoch, interval_length, coefficients = segment.load_array()
        self.assertEqual(coefficients.shape, (3, 1760, 11))

    def test_out_of_range_dates(self):
        segment = self.spk[0,4]
        tdb = np.array([-1e3, 0, +1e5]) + 2414990.0
        try:
            segment.compute_and_differentiate(tdb)
        except OutOfRangeError as e:
            self.assertEqual(str(e), 'segment only covers dates'
                             ' 1899-07-29 through 2053-10-09')
            self.assertIs(type(e.out_of_range_times), np.ndarray)
            self.assertEqual(list(e.out_of_range_times), [True, False, True])

    def test_whether_bad_ephemeris_leaves_file_open(self):
        # This doesn't actually fail if the file object is left open,
        # but should produce a ResourceWarning as a side effect.
        non_spk_path = __file__
        with self.assertRaises(ValueError):
            SPK.open(non_spk_path)

class LegacyTests(_CommonTests, TestCase):

    def setUp(self):
        try:
            import de421
        except ImportError:
            raise SkipTest('the "de421" ephemeris package has not been'
                           ' installed with "pip install de421"')
        self.eph = Ephemeris(de421)
        self.jalpha = self.eph.jalpha
        self.jomega = self.eph.jomega

    def position(self, name, tdb, tdb2=0.0):
        return self.eph.position(name, tdb, tdb2)

    def position_and_velocity(self, name, tdb, tdb2=0.0):
        return self.eph.position_and_velocity(name, tdb, tdb2)

    def test_names(self):
        self.assertEqual(self.eph.names,  (
            'earthmoon', 'jupiter', 'librations', 'mars', 'mercury',
            'moon', 'neptune', 'nutations', 'pluto', 'saturn', 'sun',
            'uranus', 'venus',
            ))

    def test_legacy_compute_method(self):
        pv = self.eph.compute('earthmoon', 2414994.0)
        self.check0(pv[:3], pv[3:])
        pv = self.eph.compute('earthmoon', np.array([2414994.0, 2415112.5]))
        self.check0(pv[:3,0], pv[3:,0])
        self.check1(pv[:3,1], pv[3:,1])

    def test_ephemeris_end_date(self):
        x, y, z = self.position('earthmoon', self.jomega)
        self.assertAlmostEqual(x, -94189805.73967789, delta=epsilon_m)
        self.assertAlmostEqual(y, 1.05103857e+08, delta=1.0)
        self.assertAlmostEqual(z, 45550861.44383482, delta=epsilon_m)

class PCKTests(TestCase):
    def test_out_of_range_date(self):
        p = PCK.open('moon_pa_de421_1900-2050.bpc')
        segment = p.segments[0]
        expect = 'segment only covers dates 1900-01-01 through 2051-01-01'
        with self.assertRaisesRegex(ValueError, expect):
            segment.compute(0.0, 0.0)
        p.close()

class NAIF_DAF_Tests(TestCase):

    def test_single_position(self):
        with SPK(NAIF_DAF(open('de405.bsp', 'rb'))) as kernel:
            x, y, z = kernel[0,4].compute(2457061.5)
            # Expect rough agreement with a DE430 position from our README:
            self.assertAlmostEqual(x, 2.05700211e+08, delta=2.0)
            self.assertAlmostEqual(y, 4.25141646e+07, delta=2.0)
            self.assertAlmostEqual(z, 1.39379183e+07, delta=2.0)

class CommandLineTests(TestCase):
    maxDiff = 9999

    def test_comment_command(self):
        output = commandline.main(['comment', 'de405.bsp'])
        self.assertEqual(output[:30], '; de405.bsp LOG FILE\n;\n; Creat')
        self.assertEqual(output[-30:], "rom Standish's DE405 memo <<<\n")

    def test_daf_command(self):
        self.assertEqual(commandline.main(['daf', 'de405.bsp']), """\
 1 DE-405 -1577879958.8160586 1577880064.1839132 1 0 1 2 1409 202316
 2 DE-405 -1577879958.8160586 1577880064.1839132 2 0 1 2 202317 275376
 3 DE-405 -1577879958.8160586 1577880064.1839132 3 0 1 2 275377 368983
 4 DE-405 -1577879958.8160586 1577880064.1839132 4 0 1 2 368984 408957
 5 DE-405 -1577879958.8160586 1577880064.1839132 5 0 1 2 408958 438653
 6 DE-405 -1577879958.8160586 1577880064.1839132 6 0 1 2 438654 464923
 7 DE-405 -1577879958.8160586 1577880064.1839132 7 0 1 2 464924 487767
 8 DE-405 -1577879958.8160586 1577880064.1839132 8 0 1 2 487768 510611
 9 DE-405 -1577879958.8160586 1577880064.1839132 9 0 1 2 510612 533455
10 DE-405 -1577879958.8160586 1577880064.1839132 10 0 1 2 533456 613364
11 DE-405 -1577879958.8160586 1577880064.1839132 301 3 1 2 613365 987780
12 DE-405 -1577879958.8160586 1577880064.1839132 399 3 1 2 987781 1362196
13 DE-405 -1577879958.8160586 1577880064.1839132 199 1 1 2 1362197 1362208
14 DE-405 -1577879958.8160586 1577880064.1839132 299 2 1 2 1362209 1362220
15 DE-405 -1577879958.8160586 1577880064.1839132 499 4 1 2 1362221 1362232
""")

    def test_spk_command(self):
        self.assertEqual(commandline.main(['spk', 'de405.bsp']), """\
File type NAIF/DAF and format BIG-IEEE with 15 segments:
1950-01-01..2050-01-01  Type 2  Solar System Barycenter (0) -> Mercury Barycenter (1)
1950-01-01..2050-01-01  Type 2  Solar System Barycenter (0) -> Venus Barycenter (2)
1950-01-01..2050-01-01  Type 2  Solar System Barycenter (0) -> Earth Barycenter (3)
1950-01-01..2050-01-01  Type 2  Solar System Barycenter (0) -> Mars Barycenter (4)
1950-01-01..2050-01-01  Type 2  Solar System Barycenter (0) -> Jupiter Barycenter (5)
1950-01-01..2050-01-01  Type 2  Solar System Barycenter (0) -> Saturn Barycenter (6)
1950-01-01..2050-01-01  Type 2  Solar System Barycenter (0) -> Uranus Barycenter (7)
1950-01-01..2050-01-01  Type 2  Solar System Barycenter (0) -> Neptune Barycenter (8)
1950-01-01..2050-01-01  Type 2  Solar System Barycenter (0) -> Pluto Barycenter (9)
1950-01-01..2050-01-01  Type 2  Solar System Barycenter (0) -> Sun (10)
1950-01-01..2050-01-01  Type 2  Earth Barycenter (3) -> Moon (301)
1950-01-01..2050-01-01  Type 2  Earth Barycenter (3) -> Earth (399)
1950-01-01..2050-01-01  Type 2  Mercury Barycenter (1) -> Mercury (199)
1950-01-01..2050-01-01  Type 2  Venus Barycenter (2) -> Venus (299)
1950-01-01..2050-01-01  Type 2  Mars Barycenter (4) -> Mars (499)
""")

    def test_verbose_spk_command_with_tidy_ephemeris(self):
        output = commandline.main(['spk', '-v', 'de421.bsp'])
        pieces = output.split('\n\n')
        pieces[1:-2] = ['...']
        output = '\n\n'.join(pieces)
        self.assertEqual(output, """\
File type DAF/SPK and format LTL-IEEE with 15 segments:
1899-07-29..2053-10-09  Type 2  Solar System Barycenter (0) -> Mercury Barycenter (1)
   7040 polynomials covering 8.0 days each
      x 14 coefficients per polynomial
      x 3 coordinates
      = 295680 double precision floats
   Polynomial start date matches segment start date
   Polynomial end date matches segment end date

...

1899-07-29..2053-10-09  Type 2  Mars Barycenter (4) -> Mars (499)
   1 polynomial covering 56320.0 days
      x 2 coefficients per polynomial
      x 3 coordinates
      = 6 double precision floats
   Polynomial start date matches segment start date
   Polynomial end date matches segment end date

""")

    def test_verbose_spk_command_with_untidy_ephemeris(self):
        output = commandline.main(['spk', '-v', 'de442s.bsp'])
        pieces = output.split('\n\n')
        pieces[1:-2] = ['...']
        output = '\n\n'.join(pieces)
        self.assertEqual(output, """\
File type DAF/SPK and format LTL-IEEE with 14 segments:
1849-12-26..2150-01-22  Type 2  Venus Barycenter (2) -> Venus (299)
   1 polynomial covering 401792.0 days
      x 2 coefficients per polynomial
      x 3 coordinates
      = 6 double precision floats
   First polynomial starts 109568.0 days earlier than segment start date, on 1549-12-31
   Final polynomial ends 182624.0 days later than segment end date, on 2650-01-25

...

1849-12-26..2150-01-22  Type 2  Solar System Barycenter (0) -> Mercury Barycenter (1)
   13701 polynomials covering 8.0 days each
      x 14 coefficients per polynomial
      x 3 coordinates
      = 575442 double precision floats
   Polynomial start date matches segment start date
   Final polynomial ends 8.0 days later than segment end date, on 2150-01-30

""")

    def test_excerpt_command(self):
        output = commandline.main(['excerpt', '2023/8/23', '2023/8/24',
                                   'de421.bsp', 'de421_excerpt.bsp'])
        self.assertEqual(output, """\
Date 2023/8/23  = JD 2460179.5
Date 2023/8/24  = JD 2460180.5

'de421_excerpt.bsp' written successfully with the following contents

File type DAF/SPK and format LTL-IEEE with 15 segments:
2023-08-23..2023-08-24  Type 2  Solar System Barycenter (0) -> Mercury Barycenter (1)
2023-08-23..2023-08-24  Type 2  Solar System Barycenter (0) -> Venus Barycenter (2)
2023-08-23..2023-08-24  Type 2  Solar System Barycenter (0) -> Earth Barycenter (3)
2023-08-23..2023-08-24  Type 2  Solar System Barycenter (0) -> Mars Barycenter (4)
2023-08-23..2023-08-24  Type 2  Solar System Barycenter (0) -> Jupiter Barycenter (5)
2023-08-23..2023-08-24  Type 2  Solar System Barycenter (0) -> Saturn Barycenter (6)
2023-08-23..2023-08-24  Type 2  Solar System Barycenter (0) -> Uranus Barycenter (7)
2023-08-23..2023-08-24  Type 2  Solar System Barycenter (0) -> Neptune Barycenter (8)
2023-08-23..2023-08-24  Type 2  Solar System Barycenter (0) -> Pluto Barycenter (9)
2023-08-23..2023-08-24  Type 2  Solar System Barycenter (0) -> Sun (10)
2023-08-23..2023-08-24  Type 2  Earth Barycenter (3) -> Moon (301)
2023-08-23..2023-08-24  Type 2  Earth Barycenter (3) -> Earth (399)
2023-08-23..2023-08-24  Type 2  Mercury Barycenter (1) -> Mercury (199)
2023-08-23..2023-08-24  Type 2  Venus Barycenter (2) -> Venus (299)
2023-08-23..2023-08-24  Type 2  Mars Barycenter (4) -> Mars (499)
""")

        preface = """\
;
; This is an ephemeris excerpt created by jplephem 2.23, which was
; asked to narrow the ephemeris to Julian dates 2460179.5 - 2460180.5
; (proleptic Gregorian dates 2023-08-23 through 2023-08-24).
;
; Here is the comments area from the original ephemeris file:
; ----------------------------------------------------------------------
"""
        output1 = commandline.main(['comment', 'de421.bsp'])
        output2 = commandline.main(['comment', 'de421_excerpt.bsp'])
        self.assertEqual(preface + output1, output2)

        output = commandline.main(['spk', '-v', 'de421_excerpt.bsp'])
        pieces = output.split('\n\n')
        pieces[1:-2] = ['...']
        output = '\n\n'.join(pieces)
        self.assertEqual(output, """\
File type DAF/SPK and format LTL-IEEE with 15 segments:
2023-08-23..2023-08-24  Type 2  Solar System Barycenter (0) -> Mercury Barycenter (1)
   1 polynomial covering 8.0 days
      x 14 coefficients per polynomial
      x 3 coordinates
      = 42 double precision floats
   First polynomial starts 3.0 days earlier than segment start date, on 2023-08-20
   Final polynomial ends 4.0 days later than segment end date, on 2023-08-28

...

2023-08-23..2023-08-24  Type 2  Mars Barycenter (4) -> Mars (499)
   1 polynomial covering 56320.0 days
      x 2 coefficients per polynomial
      x 3 coordinates
      = 6 double precision floats
   First polynomial starts 45315.0 days earlier than segment start date, on 1899-07-29
   Final polynomial ends 11004.0 days later than segment end date, on 2053-10-09

""")

def load_tests(loader, tests, ignore):
    """Run our main documentation as a test."""

    # If we are running in CI, where we test against an old version of
    # NumPy, skip the doctests since NumPy will print whitespace
    # differently (and worse).
    version = tuple(int(s) for s in np.__version__.split('.'))
    if version < (1, 17):
        return tests

    # Python 2.6 formats floating-point numbers a bit differently and
    # breaks the doctest.
    if sys.version_info <= (2, 6):
        return tests

    tests.addTests(DocTestSuite('jplephem', optionflags=ELLIPSIS))
    return tests



"""Extract data for a specific date range from an SPK file."""

from sys import stderr
try:
    from urllib.request import Request, urlopen
except ImportError:
    from urllib2 import Request, urlopen

from numpy import copy
from . import __version__ as jplephem_version
from .calendar import compute_calendar_date
from .daf import DAF, K
from .spk import S_PER_DAY, T0

clip_lower = max
clip_upper = min

_PREFACE = """\
;
; This is an ephemeris excerpt created by jplephem {}, which was
; asked to narrow the ephemeris to Julian dates {:.1f} - {:.1f}
; (proleptic Gregorian dates {}-{:02}-{:02} through {}-{:02}-{:02}).
;
; Here is the comments area from the original ephemeris file:
; ----------------------------------------------------------------------
"""

def _seconds(jd):
    """Convert a Julian Date to a number of seconds since J2000."""
    return (jd - T0) * S_PER_DAY

def write_excerpt(input_spk, output_file, start_jd, end_jd, summaries):
    start_seconds = _seconds(start_jd)
    end_seconds = _seconds(end_jd)
    old = input_spk.daf

    # Supplement the comment text.
    y1, m1, d1 = compute_calendar_date(int(start_jd + 0.5))
    y2, m2, d2 = compute_calendar_date(int(end_jd + 0.5))
    preface = _PREFACE.format(
        jplephem_version, start_jd, end_jd, y1,m1,d1, y2,m2,d2,
    )
    comment = preface + old.comments()

    # Build new comment blocks (which have 1000 text characters each).
    data = comment.encode('ascii').replace(b'\n', b'\0') + b'\004'
    blocks = [data[i : i + 1000] for i in range(0, len(data), 1000)]
    comment_data = b''.join([
        block + b' ' * (K - len(block))
        for block in blocks
    ])

    # Start the new DAF file with:
    # 1. The verbatim first record from the original file.
    # 2. The new comment.
    # 3. An empty summary block.
    # 4. An empty name block.

    f = output_file
    f.seek(0)
    f.truncate()

    summary_data = b'\0' * 1024
    name_data = b' ' * 1024

    f.write(old.read_record(1))
    f.write(comment_data)
    f.write(summary_data)
    f.write(name_data)

    # There are now enough blocks to start treating the file as a DAF!
    # Set the initial block number indexes.
    f.seek(0)
    d = DAF(f)
    d.fward = d.bward = 2 + len(comment_data) // K
    d.free = (d.fward + 1) * (1024 // 8) + 1
    d.write_file_record()

    # Copy over an excerpt of each array.
    for name, values in summaries:
        start, end = values[-2], values[-1]
        init, intlen, rsize, n = old.read_array(end - 3, end)
        rsize = int(rsize)

        i = int(clip(0, n, (start_seconds - init) // intlen))
        j = int(clip(0, n, (end_seconds - init) // intlen + 1))
        if i == j:
            continue  # Segment has no overlap with user's dates.

        init = init + i * intlen
        n = j - i

        extra = 4     # enough room to rebuild [init intlen rsize n]
        excerpt = copy(old.read_array(
            start + rsize * i,
            start + rsize * j + extra - 1,
        ))
        excerpt[-4:] = (init, intlen, rsize, n)

        # Even though the polynomials we selected probably cover a wider
        # range of dates, let's only claim that each segment covers the
        # range `start_seconds .. end_seconds` that the user requested,
        # to avoid confusing them.  The `de442s.bsp` ephemeris shows
        # that this is also the practice at NASA itself.

        values = (start_seconds, end_seconds) + values[2:]
        d.add_array(name, values, excerpt)

def clip(lower, upper, n):
    return clip_lower(lower, clip_upper(upper, n))

class RemoteFile(object):
    def __init__(self, url):
        self.url = url
        self.filename = url.rstrip('/').rsplit('/', 1)[-1]
        self.offset = 0

    def seek(self, offset, whence=0):
        assert whence == 0
        self.offset = offset

    def read(self, size):
        start = self.offset
        end = start + size - 1
        assert end > start
        byte_range = 'bytes={}-{}'.format(start, end)
        stderr.write('Fetching {} bytes from {} using Range: {}\n'
                     .format(size, self.filename, byte_range))
        request = Request(self.url, headers={'Range': byte_range})
        data = urlopen(request).read()
        assert len(data) == size, (
            'asked for "Range: {}" which is {} bytes, but got {} bytes back'
            .format(byte_range, size, len(data))
        )
        self.offset += size
        return data

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



"""The `python -m jplephem` command line."""

from __future__ import print_function

import argparse
import sys
from .calendar import compute_calendar_date, compute_julian_date
from .daf import DAF
from .excerpter import RemoteFile, write_excerpt
from .spk import S_PER_DAY, SPK, T0

_DAY = 86400.0

def _jd(seconds):
    """Convert a number of seconds since J2000 to a Julian Date."""
    return T0 + seconds / S_PER_DAY

def main(args):
    parser = argparse.ArgumentParser(
        prog='python -m jplephem',
        description='Describe an SPK kernel',
    )
    subparsers = parser.add_subparsers()

    p = subparsers.add_parser(
        'comment',
        help="Print a file's comment blocks",
    )
    p.set_defaults(func=comment)
    p.add_argument('path', help='Path to a SPICE file')

    p = subparsers.add_parser(
        'daf',
        help="List a file's raw segment descriptors",
    )
    p.set_defaults(func=daf_segments)
    p.add_argument('path', help='Path to a SPICE file')

    p = subparsers.add_parser(
        'excerpt',
        help="Create an SPK covering a narrower range of dates",
    )
    p.set_defaults(func=excerpt)
    p.add_argument('--targets', help='Comma-separated targets to include')
    p.add_argument('start_date', help='Start date yyyy/mm/dd', type=parse_date)
    p.add_argument('end_date', help='End date yyyy/mm/dd', type=parse_date)
    p.add_argument('path_or_url', help='Local filename or remote URL')
    p.add_argument('output_path', help='Output file to create')

    p = subparsers.add_parser(
        'spk',
        help="List the segments in an SPK file",
    )
    p.set_defaults(func=spk_segments)
    p.add_argument('path', help='Path to a .bsp SPICE kernel file')
    p.add_argument('-v', '--verbose', action='store_true')

    args = parser.parse_args(args)
    func = getattr(args, 'func', None)
    if func is None:
        parser.print_help()
        sys.exit(2)

    lines = list(func(args))
    if lines and not lines[-1].endswith('\n'):
        lines.append('')
    return '\n'.join(lines)

def comment(args):
    with open(args.path, 'rb') as f:
        d = DAF(f)
        yield d.comments()

def daf_segments(args):
    with open(args.path, 'rb') as f:
        d = DAF(f)
        for i, (name, values) in enumerate(d.summaries()):
            yield '{:2d} {} {}'.format(i + 1, name.decode('latin-1'),
                                       ' '.join(repr(v) for v in values))

def excerpt(args):
    for string, jd in args.start_date, args.end_date:
        yield 'Date {:10} = JD {}'.format(string, jd)

    if args.path_or_url.startswith(('http://', 'https://')):
        url = args.path_or_url
        f = RemoteFile(url)
    else:
        path = args.path_or_url
        f = open(path, 'rb')

    with f:
        spk = SPK(DAF(f))
        summaries = spk.daf.summaries()

        if args.targets:
            desired_targets = set(args.targets.split(','))
            summaries = [
                summary for summary, segment in zip(summaries, spk.segments)
                if str(segment.target) in desired_targets
            ]

        with open(args.output_path, 'w+b') as output_file:
            write_excerpt(spk, output_file, args.start_date[1],
                          args.end_date[1], summaries)

    yield '\n{!r} written successfully with the following contents\n'.format(
        args.output_path)

    with open(args.output_path, 'rb') as f:
        yield str(SPK(DAF(f)))

def spk_segments(args):
    with open(args.path, 'rb') as f:
        spk = SPK(DAF(f))

        # Snag the first line from the normal str().

        output = str(spk)
        yield output.split('\n', 1)[0]

        # But produce the rest of the lines ourselves, so we can
        # optionally honor '-v' by providing more information.

        for s in spk.segments:
            yield str(s)
            if not args.verbose:
                continue
            for line in _describe_segment_details(s):
                yield line

def _describe_segment_details(s):
    if s.data_type not in (2, 3):
        return

    init, intlen, coefficients = s._data
    degree, dimensions, record_count = coefficients.shape
    days_per = intlen / _DAY

    plural = '' if record_count == 1 else 's'
    each = '' if record_count == 1 else ' each'
    yield '   {} polynomial{} covering {} days{}'.format(
        record_count, plural, days_per, each,
    )
    yield '      x {} coefficients per polynomial'.format(degree)
    yield '      x {} coordinates'.format(dimensions)
    yield '      = {} double precision floats'.format(coefficients.size)

    polynomial_start = init
    polynomial_end = init + intlen * record_count
    if s.start_second == polynomial_start:
        yield '   Polynomial start date matches segment start date'
    else:
        days = (s.start_second - polynomial_start) / _DAY
        jd = _jd(polynomial_start)
        y, m, d = compute_calendar_date(int(jd + 0.5))
        yield (
            '   First polynomial starts {:.1f} days earlier'
             ' than segment start date, on {}-{:02}-{:02}'
             .format(days, y, m, d)
        )
    if s.end_second == polynomial_end:
        yield '   Polynomial end date matches segment end date'
    else:
        days = (polynomial_end - s.end_second) / _DAY
        jd = _jd(polynomial_end)
        y, m, d = compute_calendar_date(int(jd + 0.5))
        yield (
            '   Final polynomial ends {:.1f} days later'
            ' than segment end date, on {}-{:02}-{:02}'
            .format(days, y, m, d)
        )

    yield ''

def parse_date(s):
    try:
        fields = [int(f) for f in s.split('/')]
    except ValueError:
        fields = []
    if len(fields) < 1 or len(fields) > 3:
        E = argparse.ArgumentTypeError
        raise E('specify each date as YYYY or YYYY/MM or YYYY/MM/DD')
    jd = compute_julian_date(*fields)
    return s, jd
